@@ -1,7 +1,7 @@
-from numpy import allclose, sum, array, maximum, sqrt
 import pytest
+from numpy import allclose, array, maximum, sqrt, sum
 
-from SeismicMesh import generate_mesh, sliver_removal, geometry
+from SeismicMesh import generate_mesh, geometry, sliver_removal
 
 
 @pytest.mark.serial
@@ -39,7 +39,7 @@ def test_3dmesher_SDF():
     )
 
     points, cells = sliver_removal(
-        points=points, domain=cylinder, edge_length=EF, h0=hmin
+        points=points, domain=cylinder, edge_length=EF, h0=hmin, bbox=bbox
     )
 
     assert allclose(sum(geometry.simp_vol(points, cells)), 6.28, atol=hmin)
